/*
 * gui_lib.c
 * Copyright (c) 2002-2003 HighPoint Technologies, Inc. All rights reserved.
 *
 *  Platform independent ioctl interface implementation.
 *  The platform dependent part may reuse this function and/or use it own 
 *  implementation for each ioctl function.
 *
 *  This implementation doesn't use any synchronization; the caller must
 *  assure the proper context when calling these functions.
 */
 
#include "global.h"
#ifndef __KERNEL__
#define __KERNEL__
#endif
#include "hptintf.h"

#define INVALID_DEVICEID		(-1)
#define INVALID_STRIPSIZE		(-1)

#define shortswap(w) ((WORD)((w)>>8 | ((w) & 0xFF)<<8))

#ifndef MinBlockSizeShift
#define MinBlockSizeShift 5
#define MaxBlockSizeShift 12
#endif

#pragma pack(1)

typedef struct _HPT_IOCTL_TRANSFER_PARAM
{
	ULONG nInBufferSize;
	ULONG nOutBufferSize;
	UCHAR buffer[0];
}HPT_IOCTL_TRANSFER_PARAM, *PHPT_IOCTL_TRANSFER_PARAM;

typedef struct _HPT_SET_STATE_PARAM
{
	DEVICEID idArray;
	DWORD state;
} HPT_SET_STATE_PARAM, *PHPT_SET_STATE_PARAM;

typedef struct _HPT_SET_ARRAY_INFO
{
	DEVICEID idArray;
	ALTERABLE_ARRAY_INFO Info;
} HPT_SET_ARRAY_INFO, *PHPT_SET_ARRAY_INFO;

typedef struct _HPT_SET_DEVICE_INFO
{
	DEVICEID idDisk;
	ALTERABLE_DEVICE_INFO Info;
} HPT_SET_DEVICE_INFO, *PHPT_SET_DEVICE_INFO;

typedef struct _HPT_ADD_DISK_TO_ARRAY
{
	DEVICEID idArray;
	DEVICEID idDisk;
} HPT_ADD_DISK_TO_ARRAY, *PHPT_ADD_DISK_TO_ARRAY;

typedef struct _HPT_DEVICE_IO
{
	DEVICEID	id;
	int			cmd;
	ULONG		lba;
	DWORD		nSector;
	UCHAR		buffer[0];
} HPT_DEVICE_IO, *PHPT_DEVICE_IO;

#ifdef SUPPORT_HPT601
typedef struct _HPT_SET_601_INFO
{
	DEVICEID idDisk;
	HPT601_INFO Info;
} HPT_SET_601_INFO, *PHPT_SET_601_INFO;
#endif

#pragma pack()

static int check_VDevice_valid(PVDevice p)
{
	PVDevice pVDevice;

	for(pVDevice = pVDevStart; pVDevice < pVDevEnd; pVDevice++)
		if(pVDevice == p)	return 0;

	return -1;
}


static void get_disk_info(PVDevice pVDevice, PLOGICAL_DEVICE_INFO pInfo)
{
	PChipInstance pChip;
	PChannel pChan;
	int i;
	char *p;

	pInfo->Type = LDT_DEVICE;

	if (pVDevice->pParent)
		pInfo->ParentArray = (DEVICEID)(pVDevice->pParent);
	else
		pInfo->ParentArray = INVALID_DEVICEID;

	/* report real capacity to be compatible with old arrays */
	pInfo->Capacity = pVDevice->u.disk.dDeRealCapacity;

	/* device location */
	pChan = pVDevice->u.disk.pChannel;
	pInfo->u.device.TargetId = (&pVDevice->u.disk==pChan->pDevice[0])? 0 : 1;
	pChip = pChan->pChipInstance;
	i = pChan - pChip->pChannel[0];
	while (!pChip->first_on_slot) i+=pChip--->bCpMaxChannel;
	pInfo->u.device.PathId = i;
	i = 0;
	while (pChip>gChipInstances) if (pChip--->first_on_slot) i++;
	pInfo->u.device.ControllerId = i;

	pInfo->u.device.DeviceModeSetting = pVDevice->u.disk.bDeModeSetting;
	pInfo->u.device.UsableMode = pVDevice->u.disk.bDeUsable_Mode;
	if (pVDevice->u.disk.df_cdrom_device)
		pInfo->u.device.DeviceType = PDT_CDROM;
	else if (pVDevice->u.disk.df_tape_device)
		pInfo->u.device.DeviceType = PDT_TAPE;
	else
		pInfo->u.device.DeviceType = PDT_HARDDISK;

	pInfo->u.device.Flags = 0x0;

	/* device is disabled */
	if(!pVDevice->u.disk.df_on_line)
		pInfo->u.device.Flags |= DEVICE_FLAG_DISABLED;

#ifdef SUPPORT_HPT601
    if (pVDevice->u.disk.df_with_601)
		pInfo->u.device.Flags |= DEVICE_FLAG_WITH_601;
#endif

	/* disk has a active partition */
	if(pVDevice->vf_bootable)
		pInfo->u.device.Flags |= DEVICE_FLAG_BOOTDISK;

	/* disk has boot mark set */
	if(pVDevice->vf_bootmark)
		pInfo->u.device.Flags |= DEVICE_FLAG_BOOTMARK;
		
	if (pVDevice->u.disk.df_sata)
		pInfo->u.device.Flags |= DEVICE_FLAG_SATA;

	/* is a spare disk */
	if(pVDevice->VDeviceType == VD_SPARE)
		pInfo->u.device.Flags |= DEVICE_FLAG_IS_SPARE;

	memcpy(&(pInfo->u.device.IdentifyData), &(pVDevice->u.disk.IdentifyData), sizeof(IDENTIFY_DATA2));
	p = (char *)&pInfo->u.device.IdentifyData.ModelNumber;
	for (i = 0; i < 20; i++)
		((WORD*)p)[i] = shortswap(pVDevice->u.disk.IdentifyData.ModelNumber[i]);
	p[39] = '\0';
}

int hpt_get_driver_capabilities(PDRIVER_CAPABILITIES cap)
{
	ZeroMemory(cap, sizeof(DRIVER_CAPABILITIES));
	cap->dwSize = sizeof(DRIVER_CAPABILITIES);
	cap->MaximumControllers = MAX_CHIPS;

	/* cap->SupportCrossControllerRAID = 0; */
	/* take care for various OSes! */

	cap->MinimumBlockSizeShift = MinBlockSizeShift;
	cap->MaximumBlockSizeShift = MaxBlockSizeShift;
	cap->SupportDiskModeSetting = 0;
	cap->SupportSparePool = 1;		
	cap->MaximumArrayNameLength = MAX_ARRAY_NAME - 1;
	cap->SupportDedicatedSpare = 0;
	
#ifdef SUPPORT_RAID15
	cap->HighPerformanceRAID1 = 1;
#endif


	return 0;
}

int hpt_get_controller_count(void)
{
	PChipInstance pChip;
	int iControllerCount = 0;

	for (pChip = gChipInstances; pChip<pChipEnd; pChip++)
		if (pChip->first_on_slot)
			iControllerCount++;
	return iControllerCount;
}

int hpt_get_controller_info(int id, PCONTROLLER_INFO pInfo)
{
	PChipInstance pChip;
	int iControllerCount = 0;

	for (pChip = gChipInstances; pChip<pChipEnd; pChip++) {
		if (!pChip->first_on_slot) continue;
		if (iControllerCount++==id) {
		    pInfo->InterruptLevel = pChip->bChipIntrNum;
			pInfo->ChipType = pChip->bChipType;
			pInfo->ChipFlags = CHIP_SUPPORT_ULTRA_100;
		    strcpy( pInfo->szVendorID, "HighPoint Technologies, Inc.");
#ifdef GUI_CONTROLLER_NAME
#ifdef FORCE_ATA150_DISPLAY
			/* show "Bus Type: ATA/150" in GUI for S-ATA controllers */
			pInfo->ChipFlags = CHIP_SUPPORT_ULTRA_150;
#endif
			strcpy(pInfo->szProductID, GUI_CONTROLLER_NAME);
#define _set_product_id(x)
#else 
#define _set_product_id(x) strcpy(pInfo->szProductID, x)
#endif
			switch(pChip->bChipType) {
			case CHIP_TYPE_HPT374:
				_set_product_id("HPT374 UDMA/ATA133 RAID Controller");
				pInfo->NumBuses = 4;
				pInfo->ChipFlags |= CHIP_SUPPORT_ULTRA_133;
				return 0;
			case CHIP_TYPE_HPT371:
			case CHIP_TYPE_HPT371N:
				_set_product_id("HPT371 UDMA/ATA133 RAID Controller");
				pInfo->NumBuses = 1;
				pInfo->ChipFlags |= CHIP_SUPPORT_ULTRA_133;
				return 0;
			/* add other chip types if needed */
			case CHIP_TYPE_HPT372:
			case CHIP_TYPE_HPT372A:
			case CHIP_TYPE_HPT372N:
			case CHIP_TYPE_HPT302:
			case CHIP_TYPE_HPT302N:
				pInfo->ChipFlags |= CHIP_SUPPORT_ULTRA_133;
				/* fall down */
			default:
				_set_product_id("HPT3xx ATA RAID Controller");
				pInfo->NumBuses = 2;
				return 0;
			}
		}
	}
	return -1;
}

int hpt_get_channel_info(int id, int bus, PCHANNEL_INFO pInfo)
{
	PChipInstance pChip;
	int iControllerCount = 0;
	int iDev;

	for (pChip = gChipInstances; pChip<pChipEnd; pChip++) {
		if (!pChip->first_on_slot) continue;
		if (iControllerCount++==id) goto found;
	}
	return -1;

found:
	/* the channel maybe on other functions of the controller. */
	while (bus>=pChip->bCpMaxChannel) {
		bus -= pChip->bCpMaxChannel;
		pChip++;
		if (pChip>=pChipEnd || pChip->first_on_slot) return -1;
	}

	pInfo->IoPort = (ULONG)pChip->pChannel[bus]->BaseIoAddress1;
	pInfo->ControlPort = (ULONG)pChip->pChannel[bus]->BaseIoAddress2;

	for(iDev = 0; iDev < MAX_DEVICE_PER_CHANNEL; iDev++)
	{
		if (pChip->pChannel[bus]->pDevice[iDev])
			pInfo->Devices[iDev] = (DEVICEID)Map2pVDevice(pChip->pChannel[bus]->pDevice[iDev]);
		else
			pInfo->Devices[iDev] = (DEVICEID)INVALID_DEVICEID;
	}
	return 0;
}

int hpt_get_logical_devices(DEVICEID * pIds, int nMaxCount)
{
	int count = 0;
	int	i;
    PVDevice pPhysical, pLogical;
	PVBus pVBus;

	for(i = 0; i < nMaxCount; i++)
		pIds[i] = INVALID_DEVICEID;

	/* first copy the registered devices */
	for (pVBus = gVBus; pVBus<&gVBus[MAX_VBUS]; pVBus++) {
		for(i = 0; i < MAX_VDEVICE_PER_VBUS; i++) {
			if ((pIds[count] = (DEVICEID)pVBus->pVDevice[i])) count++;
			if (count>=nMaxCount) goto done;
		}
	}

	/* append the arrays not registered on VBus */
	for(pPhysical = pVDevStart; pPhysical < pVDevEnd; pPhysical++) {
		pLogical = pPhysical;
		while (pLogical->pParent) pLogical = pLogical->pParent;
		if (pLogical->VDeviceType==VD_SPARE)
			continue;
		for (i=0; i<count; i++)
			if (pIds[i]==(DEVICEID)pLogical) goto next;
		pIds[count++] = (DEVICEID)pLogical;
		if (count>=nMaxCount) goto done;
		next:;
	}
done:
	return count;
}

int hpt_get_device_info(DEVICEID id, PLOGICAL_DEVICE_INFO pInfo)
{
	PVDevice pVDevice = (PVDevice)id;

	if((pVDevice == NULL) || check_VDevice_valid(pVDevice))
		return -1;

		get_disk_info(pVDevice, pInfo);

	return 0;
}


#ifdef SUPPORT_HPT601
int hpt_get_601_info(DEVICEID idDisk, PHPT601_INFO pInfo)
{
	PVDevice pVDevice = (PVDevice)idDisk;
    PChannel          pChan = pVDevice->u.disk.pChannel;
    PIDE_REGISTERS_1  IoPort = pChan->BaseIoAddress1;

    if(!pVDevice->u.disk.df_with_601) return -1;

	mSelectUnit(IoPort, pVDevice->u.disk.bDeUnitId);
	pChan->pChipInstance->ftbl.pfnWaitOnBusy(pChan, pVDevice->u.disk.bDeUnitId, 1);
	
	BeginAccess601(IoPort);

	mSetBlockCount(IoPort, 0);
	pInfo->DeviceId = InWord(&IoPort->Data);

	mSetBlockCount(IoPort, 0x14);
	pInfo->Temperature = InWord(&IoPort->Data);

	mSetBlockCount(IoPort, 0xA);
	pInfo->FanStatus = InWord(&IoPort->Data);

	mSetBlockCount(IoPort, 7);
	pInfo->BeeperControl = InWord(&IoPort->Data);

	mSetBlockCount(IoPort, 3);
	pInfo->LED1Control = InWord(&IoPort->Data);

	mSetBlockCount(IoPort, 5);
	pInfo->LED2Control = InWord(&IoPort->Data);

	mSetBlockCount(IoPort, 0x18);
	pInfo->PowerStatus = InWord(&IoPort->Data);
		
	EndAccess601(IoPort);
	pInfo->ValidFields = 0x7F;
	/*DEVICEID|TEMPERATURE|FANSTATUS|BEEPERCONTROL|LED1CONTROL|LED2CONTROL|POWERSTATUS*/
	return 0;
}

int hpt_set_601_info(DEVICEID idDisk, PHPT601_INFO pInfo)
{
	PVDevice pVDevice = (PVDevice)idDisk;
	PChannel          pChan = pVDevice->u.disk.pChannel;
    PIDE_REGISTERS_1  IoPort = pChan->BaseIoAddress1;

	if(!pVDevice->u.disk.df_with_601) return -1;

    mSelectUnit(IoPort, pVDevice->u.disk.bDeUnitId);
	pChan->pChipInstance->ftbl.pfnWaitOnBusy(pChan, pVDevice->u.disk.bDeUnitId, 1);
	
	BeginAccess601(IoPort);

	if (pInfo->ValidFields & HPT601_INFO_TEMPERATURE) {
		mSetBlockCount(IoPort, 1);
		OutWord(&IoPort->Data, pInfo->Temperature);
	}

	if (pInfo->ValidFields & HPT601_INFO_FANSTATUS) {
		mSetBlockCount(IoPort, 0xA);
		OutWord(&IoPort->Data, pInfo->FanStatus);
	}

	if (pInfo->ValidFields & HPT601_INFO_BEEPERCONTROL) {
		mSetBlockCount(IoPort, 7);
		OutWord(&IoPort->Data, pInfo->BeeperControl);
	}

	if (pInfo->ValidFields & HPT601_INFO_LED1CONTROL) {
		mSetBlockCount(IoPort, 3);
		OutWord(&IoPort->Data, pInfo->LED1Control);
	}

	if (pInfo->ValidFields & HPT601_INFO_LED2CONTROL) {
		mSetBlockCount(IoPort, 5);
		OutWord(&IoPort->Data, pInfo->LED2Control);
	}
		
	EndAccess601(IoPort);

	return 0;
}
#endif

/* hpt_default_ioctl()
 *  This is a default implementation. The platform dependent part
 *  may reuse this function and/or use it own implementation for
 *  each ioctl function.
 */
int hpt_default_ioctl(_VBUS_ARG
						DWORD dwIoControlCode,       	/* operation control code */
						PVOID lpInBuffer,            	/* input data buffer */
						DWORD nInBufferSize,         	/* size of input data buffer */
						PVOID lpOutBuffer,           	/* output data buffer */
						DWORD nOutBufferSize,        	/* size of output data buffer */
						PDWORD lpBytesReturned      	/* byte count */
					)
{
	switch(dwIoControlCode)	{

	case HPT_IOCTL_GET_VERSION:

		if (nInBufferSize != 0) return -1;
		if (nOutBufferSize != sizeof(DWORD)) return -1;
		*((DWORD*)lpOutBuffer) = HPT_INTERFACE_VERSION;
		break;

	case HPT_IOCTL_GET_CONTROLLER_COUNT:

		if (nOutBufferSize!=sizeof(DWORD)) return -1;
		*(PDWORD)lpOutBuffer = hpt_get_controller_count();
		break;

	case HPT_IOCTL_GET_CONTROLLER_INFO:
		{
			int id;
			PCONTROLLER_INFO pInfo;
	
			if (nInBufferSize!=sizeof(DWORD)) return -1;
			if (nOutBufferSize!=sizeof(CONTROLLER_INFO)) return -1;
	
			id = *(DWORD *)lpInBuffer;
			pInfo = (PCONTROLLER_INFO)lpOutBuffer;
			if (hpt_get_controller_info(id, pInfo)!=0)
				return -1;
		}
		break;

	case HPT_IOCTL_GET_CHANNEL_INFO:
		{
			int id, bus;
			PCHANNEL_INFO pInfo;

			if (nInBufferSize!=8) return -1;
			if (nOutBufferSize!=sizeof(CHANNEL_INFO)) return -1;

			id = *(DWORD *)lpInBuffer;
			bus = ((DWORD *)lpInBuffer)[1];
			pInfo = (PCHANNEL_INFO)lpOutBuffer;

			if (hpt_get_channel_info(id, bus, pInfo)!=0)
				return -1;
		}
		break;

	case HPT_IOCTL_GET_LOGICAL_DEVICES:
		{
			DWORD nMax;
			DEVICEID *pIds;

			if (nInBufferSize!=sizeof(DWORD)) return -1;
			nMax = *(DWORD *)lpInBuffer;
			if (nOutBufferSize < sizeof(DWORD)+sizeof(DWORD)*nMax) return -1;

			pIds = ((DEVICEID *)lpOutBuffer)+1;
			*(DWORD*)lpOutBuffer = hpt_get_logical_devices(pIds, nMax);
		}
		break;

	case HPT_IOCTL_GET_DEVICE_INFO:
		{
			DEVICEID id;
			PLOGICAL_DEVICE_INFO pInfo;

			if (nInBufferSize!=sizeof(DEVICEID)) return -1;
			if (nOutBufferSize!=sizeof(LOGICAL_DEVICE_INFO)) return -1;

			id = *(DWORD *)lpInBuffer;
			if (id == INVALID_DEVICEID)	return -1;

			pInfo = (PLOGICAL_DEVICE_INFO)lpOutBuffer;
			memset(pInfo, 0, sizeof(LOGICAL_DEVICE_INFO));

			if (hpt_get_device_info(id, pInfo)!=0)
				return -1;
		}
		break;

	case HPT_IOCTL_RESCAN_DEVICES:
		{
		    void fRescanAllDevice(_VBUS_ARG0);
			if (nInBufferSize!=0) return -1;
			if (nOutBufferSize!=0) return -1;
    		fRescanAllDevice(_VBUS_P0);
		}
		break;

	case HPT_IOCTL_GET_DRIVER_CAPABILITIES:
		{
			PDRIVER_CAPABILITIES cap;
			if (nOutBufferSize<sizeof(DRIVER_CAPABILITIES)) return -1;
			cap = (PDRIVER_CAPABILITIES)lpOutBuffer;

			if(hpt_get_driver_capabilities(cap))
				return -1;
		}
		break;

#ifdef SUPPORT_HPT601	
	case HPT_IOCTL_GET_601_INFO:
		{
			DEVICEID id;
			PHPT601_INFO pInfo;

			if (nInBufferSize!=sizeof(DEVICEID)) return -1;
			if (nOutBufferSize!=sizeof(HPT601_INFO)) return -1;

			id = *(DWORD *)lpInBuffer;
			if (id == INVALID_DEVICEID)	return -1;

			pInfo = (PHPT601_INFO)lpOutBuffer;
			memset(pInfo, 0, sizeof(HPT601_INFO));

			if (hpt_get_601_info(id, pInfo)!=0)
				return -1;
		}
	    break;

    case HPT_IOCTL_SET_601_INFO:
		{
		    DEVICEID id;
			PHPT601_INFO pInfo;

			if (nInBufferSize!=sizeof(HPT_SET_601_INFO)) return -1;
			if (nOutBufferSize!=0) return -1;

			id = ((PHPT_SET_601_INFO)lpInBuffer)->idDisk;
			pInfo = &((PHPT_SET_601_INFO)lpInBuffer)->Info;
			if(hpt_set_601_info(id, pInfo) != 0)
				return -1;
		}
		break;
#endif
	default:
		return -1;
	}

	if (lpBytesReturned)
		*lpBytesReturned = nOutBufferSize;
	return 0;
}
